import { writable, derived } from "svelte/store";
const isDeprecatedLazyRoute = (obj) => typeof obj.lazyLoad === "object";
const isStaticRoute = (obj) => (typeof obj.component === "function" || typeof obj.component === "string") &&
    !obj.lazyLoad;
const isMetadataRoute = (obj) => obj.metadata === true;
const isLayoutRoute = (obj) => typeof obj.layout === "number" || obj.layout === true;
const isProtectedRoute = (obj) => typeof obj.authenticator === "function";
const defaultValue = {
    pathname: "",
    params: {},
    authStatus: "none",
    loadingStatus: "none",
    component: undefined,
    layout: undefined,
    error: undefined,
    targetName: undefined,
    // a placeholder
    // @ts-ignore
    currentRoute: { path: "", component: undefined },
    routes: [],
};
function convertLazy(oldLazyRoute, noWarn = false) {
    if (!noWarn)
        console.warn(`super-svelte-router: passed deprecated formatting for LazyLoaded routes support for this format will be dropped by v2.0.0`);
    return {
        path: oldLazyRoute.path,
        lazyLoad: true,
        // @ts-ignore
        component: oldLazyRoute.lazyLoad.component,
        loading: oldLazyRoute.lazyLoad.loading,
    };
}
function convertDeprecated(route) {
    if (isDeprecatedLazyRoute(route)) {
        if (isProtectedRoute(route))
            return Object.assign(Object.assign({}, convertLazy(route)), { authenticator: route.authenticator, authComponent: route.authComponent, authRedirect: route.authRedirect });
        return convertLazy(route);
    }
    return route;
}
function routerStoreCreator() {
    const { subscribe, update } = writable(defaultValue);
    return {
        subscribe,
        update(newValue) {
            update((value) => (Object.assign(Object.assign({}, value), newValue)));
        },
        renderRoute(route, metadata) {
            if (isStaticRoute(route)) {
                return typeof route.component === "string"
                    ? { targetName: route.component }
                    : { component: route.component };
            }
            route
                .component()
                .then((md) => this.update({ component: md.default, targetName: undefined }))
                .catch((error) => {
                this.update({ error, loadingStatus: "error" });
                console.error(error);
            });
            return {
                component: route.loading === false || typeof route.loading === "string"
                    ? undefined
                    : route.loading || (metadata === null || metadata === void 0 ? void 0 : metadata.defaultLoading),
                targetName: typeof route.loading === "string" ? route.loading : undefined,
                loadingStatus: "pending",
            };
        },
        render(value, pathname) {
            let component;
            let targetName;
            let error;
            let authStatus = "none";
            let loadingStatus = "none";
            let allowShowing = true;
            const [route, params, metadata, layout] = getRoute(pathname, value.routes);
            if (!route) {
                return Object.assign(Object.assign({}, value), { pathname });
            }
            if (isProtectedRoute(route)) {
                const authResult = route.authenticator();
                if (authResult instanceof Promise) {
                    allowShowing = false;
                    authStatus = "pending";
                    route.authComponent !== false &&
                        (component = route.authComponent || (metadata === null || metadata === void 0 ? void 0 : metadata.defaultAuthComponent));
                    authResult
                        .then((authResult) => {
                        if (route.authRedirect && !authResult) {
                            this.redirect(route.authRedirect);
                            return;
                        }
                        this.update(Object.assign({ authStatus: authResult ? "none" : "fail" }, (authResult && this.renderRoute(route, metadata))));
                    })
                        .catch((error) => {
                        console.error(error);
                        route.authRedirect
                            ? this.redirect(route.authRedirect)
                            : this.update({ error, authStatus: "error" });
                    });
                }
                else {
                    if (route.authRedirect && !authResult) {
                        updateHistory(route.authRedirect, false);
                        return this.render(value, route.authRedirect);
                    }
                    allowShowing = !!authResult;
                    authStatus = authResult ? "none" : "fail";
                    !authResult &&
                        route.authComponent !== false &&
                        (component = route.authComponent);
                }
            }
            if (typeof component === "string") {
                targetName = component;
                component = undefined;
            }
            return Object.assign({ routes: value.routes, component,
                // @ts-ignore
                targetName,
                error,
                authStatus,
                // @ts-ignore
                loadingStatus,
                pathname,
                params, currentRoute: route, layout: layout === null || layout === void 0 ? void 0 : layout.component }, (allowShowing && this.renderRoute(route, metadata)));
        },
        redirect(pathname, replace) {
            updateHistory(pathname, replace);
            update((value) => this.render(value, pathname));
        },
        setRoutes(routes, pathname) {
            update((value) => this.render(Object.assign(Object.assign({}, value), { routes: routes.map(convertDeprecated) }), pathname));
        },
    };
}
function updateHistory(pathname, replace) {
    replace
        ? history.replaceState(undefined, document.title, pathname)
        : history.pushState(undefined, document.title, pathname);
}
export const routerStore = routerStoreCreator();
export const isActive = derived(routerStore, ($routerStore) => (path) => $routerStore.currentRoute.path === path);
if (typeof window !== "undefined") {
    window.addEventListener("popstate", () => {
        routerStore.redirect(window.location.pathname);
    });
}
function getRoute(pathname, routes) {
    let params = {};
    let notFoundRoute;
    let metadataRoute;
    let layoutRoute;
    let layoutRouteFor;
    const pathnameParts = pathname.split("/");
    for (let i = 0; i < routes.length; i++) {
        const route = routes[i];
        if (isLayoutRoute(route)) {
            layoutRoute = route;
            typeof route.layout === "number" && (layoutRouteFor = route.layout);
            continue;
        }
        if (isMetadataRoute(route)) {
            metadataRoute = route;
            continue;
        }
        params = {};
        if (!isProtectedRoute(route) && route.path === "**") {
            notFoundRoute = route;
            continue;
        }
        let matches = true;
        const routeParts = route.path.split("/");
        for (const [routePart, pathnamePart] of LO(routeParts, pathnameParts)) {
            if ((pathnamePart === undefined && routePart === "") ||
                (routePart === undefined && pathnamePart === ""))
                continue;
            if (routePart === null || routePart === void 0 ? void 0 : routePart.startsWith(":")) {
                params[routePart.slice(1)] = decodeURI(pathnamePart || "");
                continue;
            }
            if (routePart !== pathnamePart) {
                matches = false;
                break;
            }
        }
        if (matches) {
            return [route, params, metadataRoute, layoutRoute];
        }
        if (layoutRoute && layoutRouteFor !== undefined) {
            layoutRouteFor--;
            if (layoutRouteFor === 0) {
                layoutRoute = undefined;
                layoutRouteFor = undefined;
            }
        }
    }
    return [notFoundRoute, params, metadataRoute, layoutRoute];
}
/** LO stands for LoopOver */
export function* LO(...args) {
    const len = Math.max.apply(undefined, args.map((arg) => arg.length));
    for (let i = 0; i < len; i++) {
        const result = [];
        for (let y = 0; y < args.length; y++) {
            result.push(args[y][i]);
        }
        yield result;
    }
}
